module Kickstand
  class EnqueuerWorker < ::ActiveJob::Base

    def perform(workflow_id)
      @workflow = Kickstand::Workflow.where(id: workflow_id).first
      return if @workflow.nil? || @workflow.stopped?

      load_jobs

      if finished?
        configuration.on_workflow_finish.call(@workflow) if configuration.on_workflow_finish
        if @workflow.auto_destroy
          configuration.on_workflow_auto_destroy.call(@workflow) if configuration.on_workflow_auto_destroy
          Kickstand::Job.where(workflow_id: @workflow.id).destroy_all #avoids having to destroy one by one
          @workflow.destroy!
        else
          next_status = @failed_jobs.any? ? "failed" : "succeeded"
          @workflow.set(finished_at: Time.now, status: next_status)
        end
        return
      end

      begin
        redis_client.secure_enqueuer_locker(workflow_id, @workflow.enqueuer_frequency)

        can_still_proceed = @failed_jobs.all? do |job|
          job.allowed_to_fail || job.retriable?
        end

        unless can_still_proceed
          @workflow.set(failed_at: Time.now, status: "failed")
          configuration.on_workflow_failure.call(@workflow) if configuration.on_workflow_failure
          return
        end

        jobs_to_be_enqueued = []

        @failed_jobs.each do |job|
          if job.retriable?
            configuration.on_job_retry.call(@workflow, job) if configuration.on_job_retry
            jobs_to_be_enqueued << job
          end
        end

        @timeout_jobs.each do |job|
          configuration.on_job_timeout.call(@workflow, job) if configuration.on_job_timeout
          jobs_to_be_enqueued << job
        end

        @lost_jobs.each do |job|
          configuration.on_job_lost.call(@workflow, job) if configuration.on_job_lost
          jobs_to_be_enqueued << job
        end

        jobs_already_being_processed = @jobs_in_progress + jobs_to_be_enqueued.size
        number_of_jobs_jobs_to_be_enqueued = @workflow.parallel_execs - jobs_already_being_processed

        counter = 0
        until (counter >= number_of_jobs_jobs_to_be_enqueued) || @ready_to_start_jobs.empty?
          job = @ready_to_start_jobs.shift
          jobs_to_be_enqueued << job
          counter = counter + 1
        end

        if jobs_to_be_enqueued.any?
          enqueue_jobs(jobs_to_be_enqueued)
        end

        redis_client.release_enqueuer_locker(workflow_id)
        Kickstand::EnqueuerWorker.set(queue: configuration.namespace, wait: @workflow.enqueuer_frequency).perform_later(workflow_id)
      rescue Kickstand::RedisClient::EnqueuerAlreadyBeingProcessed
        # Already being executed, no need to do anything
      end
    end

    private

    def enqueue_jobs(jobs)
      job_ids = jobs.map { |job| job[:id] }
      Kickstand::Job.in(_id: job_ids).set(
        enqueued_at: Time.now, started_at: nil, finished_at: nil, 
        failed_at: nil, status: "processing", updated_at: Time.now
      )
      default_queue = configuration.namespace
      workflow_id = @workflow.id.to_s
      jobs.each do |job|
        queue = job.queue || default_queue
        Kickstand::Worker.set(queue: queue).perform_later(workflow_id, job[:id].to_s)
      end
    end

    def load_jobs
      @jobs_by_id = {}
      @loaded_jobs = []
      @failed_jobs = []
      @timeout_jobs = []
      @lost_jobs = []
      @finished_jobs = []
      @processing_jobs = []
      @ready_to_start_jobs = []
      @jobs_in_progress = 0

      Kickstand::Job.where(workflow_id: @workflow.id).only(:id, :finished_at, :failed_at, :started_at, :enqueued_at, :allowed_to_fail, :timeout, :created_at, :retries, :incoming, :queue, :number_of_executions, :_type).each do |job|
        @jobs_by_id[job.id] = job
        @loaded_jobs << job
      end

      @loaded_jobs.each do |job|
        if job.incoming.any?
          incoming_jobs = @jobs_by_id.values_at(*job.incoming)
          job.load_parent_jobs(incoming_jobs)
        end

        if job.finished? || job_allowed_to_fail?(job)
          @finished_jobs << job
          next
        end

        if job.failed?
          @failed_jobs << job
          next
        end

        if job.started?
          if job.timed_out?
            @timeout_jobs << job
          else
            @processing_jobs << job
            @jobs_in_progress = @jobs_in_progress + 1
          end
          next
        end

        if job.enqueued?
          if job.lost?
            @lost_jobs << job
          else
            @processing_jobs << job
            @jobs_in_progress = @jobs_in_progress + 1
          end
          next
        end

        if job.ready_to_start?
          @ready_to_start_jobs << job
        end

      end
    end

    def finished?
      @lost_jobs.empty? && @failed_jobs.empty? && @timeout_jobs.empty? && @finished_jobs.size == @workflow.jobs.size
    end

    def job_allowed_to_fail?(job)
      job.allowed_to_fail && !job.retriable?
    end

    def current_timestamp
      @current_timestamp ||= Time.now
    end

    def configuration
      @configuration ||= Kickstand.configuration
    end

    def redis_client
      @redis_client ||= Kickstand::RedisClient.new
    end
  end
end
