module Kickstand
  class Job
    include Mongoid::Document
    include Mongoid::Timestamps
    
    prepend Kickstand::Overrides::SetWithTimestamp

    store_in collection: "kickstand_jobs"

    index({workflow_id: 1}, background: true)

    field :workflow_id, type: BSON::ObjectId
    field :incoming, type: Array, default: []
    field :outgoing, type: Array, default: [] # TODO check if we can remove this
    field :params, type: String, default: {}
    field :number_of_executions, type: Integer, default: 0
    field :timeout, type: Integer, default: 300
    field :retries, type: Integer, default: 1
    field :finished_at, type: Integer
    field :failed_at, type: Integer
    field :started_at, type: Integer
    field :enqueued_at, type: Integer
    field :queue, type: String
    field :include_payload, type: Boolean, default: true
    field :allowed_to_fail, type: Boolean, default: false
    field :output_payload, type: Hash
    field :status, type: String, default: "pending"

    belongs_to :workflow, class_name: "Kickstand::Workflow"

    def name
      @name ||= "#{self.class.name}|#{id}"
    end

    def output(data)
      self.output_payload = data
    end

    def perform
    end

    def start!
      self.started_at = current_timestamp
      self.number_of_executions = self.number_of_executions + 1
      self.failed_at = nil
    end

    def enqueue!
      self.enqueued_at = current_timestamp
      self.started_at = nil
      self.finished_at = nil
      self.failed_at = nil
    end

    def finish!
      self.finished_at = current_timestamp
    end

    def fail!
      self.finished_at = self.failed_at = current_timestamp
    end

    def enqueued?
      !enqueued_at.nil?
    end

    def lost?
      enqueued? && !started? && !finished? && !failed? && (current_timestamp > (enqueued_at + 1.hour))
    end

    def finished?
      !finished_at.nil?
    end

    def failed?
      !failed_at.nil?
    end

    def retriable?
      retries >= number_of_executions
    end

    def succeeded?
      finished? && !failed?
    end

    def started?
      !started_at.nil?
    end

    def running?
      started? && !finished? && !failed?
    end

    def timed_out?
      started? && !finished? && !failed? && (current_timestamp > (started_at + timeout.seconds))
    end

    def processing?
      running? || (enqueued? && !finished? && !failed?)
    end

    def ready_to_start?
      !running? && !enqueued? && !finished? && !failed? && parents_succeeded?
    end

    def status
      case
        when succeeded?
          :succeeded   
        when failed?
          :failed
        when running?
          :running
        when finished?
          :finished
        when processing?
          :processing
        else
          :pending
      end
    end

    def parent_jobs
      @parent_jobs ||= Kickstand::Job.in(id: incoming).to_a
    end

    def params=(data)
      encoded = Kickstand::JSON.encode(data)
      super(encoded)
    end

    def params
      data = super
      Kickstand::JSON.decode(data, symbolize_keys: true)
    end

    def load_parent_jobs(parent_jobs)
      @parent_jobs = parent_jobs
    end

    def payloads
      @payloads ||= parent_jobs.map do |job|
        {
          id: job.id,
          status: job.status,
          class: job.class.to_s,
          output: job.output_payload
        }
      end
    end
    def parents_succeeded?
      return true if has_no_dependencies?
      parent_jobs.all? do |job|
        job.succeeded? || ( job.allowed_to_fail && job.failed? && !job.retriable? )
      end
    end

    def has_no_dependencies?
      incoming.empty?
    end

    private

    def current_timestamp
      Time.now.to_i
    end

  end
end
