module Kickstand
  class Workflow
    include Mongoid::Document
    include Mongoid::Timestamps

    prepend Kickstand::Overrides::SetWithTimestamp

    store_in collection: "kickstand_workflows"

    field :stopped_at, type: DateTime
    field :started_at, type: DateTime
    field :finished_at, type: DateTime
    field :failed_at, type: DateTime
    field :starts_at, type: DateTime
    field :enqueuer_frequency, type: Integer
    field :parallel_execs, type: Integer, default: 500
    field :auto_destroy, type: Boolean, default: true
    field :status, type: String, default: "pending"
    field :tenant_id, type: BSON::ObjectId
    field :tenant_type, type: String

    index({ starts_at: 1, started_at: 1 }, background: true, sparse: true)

    has_many :jobs, class_name: "Kickstand::Job"

    def self.create(*args, delayed_enqueueing: nil, enqueuer_frequency: Kickstand.configuration.enqueuer_frequency, parallel_execs: 500, auto_destroy: false, starts_at: nil)
      flow = new(enqueuer_frequency: enqueuer_frequency, parallel_execs: parallel_execs, auto_destroy: auto_destroy, starts_at: starts_at)
      flow.setup(args)
      flow.persist_jobs!
      flow.save!
      flow
    end

    def self.delayed_create(args, enqueuer_frequency: Kickstand.configuration.enqueuer_frequency, parallel_execs: 500, auto_destroy: false, starts_at: nil)
      marshalled_args = Kickstand::JSON.encode(args)
      configuration = Kickstand.configuration
      DelayedWorkflowCreatorWorker.set(queue: configuration.namespace).perform_later(self.name, marshalled_args, enqueuer_frequency, parallel_execs, auto_destroy, starts_at)
    end

    def configure(*args)
    end

    def start!
      mark_as_started
      Kickstand::EnqueuerWorker.set(queue: Kickstand.configuration.namespace).perform_later(self.id.to_s)
    end

    def mark_as_stopped
      self.stopped_at = DateTime.now
    end

    def mark_as_failed
      self.failed_at = DateTime.now
    end

    def mark_as_started
      set(started_at: DateTime.now, stopped_at: nil, failed_at: nil, finished_at: nil, status: "running")
    end

    def link_tenant(tenant_type, tenant_id)
      self.tenant_id = tenant_id
      self.tenant_type = tenant_type&.to_s
    end

    def tenant
      return if tenant_id.blank?
      resource = tenant_type.blank? ? "Tenant" : tenant_type.capitalize
      "#{resource}: #{tenant_id}"
    end

    # only used in tests
    def find_job(id_or_name)
      Kickstand::Job.where(workflow_id: id).any_of({id: id_or_name}, {_type: id_or_name}).first
    end

    def succeeded?
      return false unless finished?
      return false if failed?
      jobs.all?(&:succeeded?)
    end

    def finished?
      !!finished_at
    end

    def started?
      !!started_at
    end

    def running?
      started? && !failed? && !stopped? && !finished?
    end

    def failed?
      !!failed_at
    end

    def stopped?
      !!stopped_at
    end

    def scheduled?
      starts_at.present?
    end

    def schedule_overdue?
      !scheduled? || starts_at <= Time.now
    end

    def run(klass, opts = {})
      @dependencies ||= []

      job = klass.new(
        workflow_id: id,
        id: BSON::ObjectId.new,
        queue: opts[:queue]
      )

      job.assign_attributes(
        opts.slice(
          :params,
          :include_payload,
          :allowed_to_fail,
          :retries,
          :timeout
        )
      )
      load_job(job)

      deps_after = [*opts[:after]]

      deps_after.each do |dep|
        @dependencies << {from: dep, to: job.id }
      end

      deps_before = [*opts[:before]]

      deps_before.each do |dep|
        @dependencies << {from: job.id, to: dep }
      end

      job.id
    end

    def reload
      self.class.find(id)
    end

    def self.descendants
      ObjectSpace.each_object(Class).select { |klass| klass < self }
    end

    def setup(args)
      configure(*args)
      resolve_dependencies
    end

    def persist_jobs!
      return unless @loaded_jobs
      now = Time.now
      job_documents = @loaded_jobs.map do |job|
        job.as_document.merge(created_at: now)
      end
      Kickstand::Job.collection.insert_many(job_documents)
    end

    private

    def load_job(job)
      @loaded_jobs ||= []
      @jobs_by_id ||= {}
      @jobs_by_name ||= {}

      @loaded_jobs << job
      @jobs_by_id[job.id] = job
      @jobs_by_name[job.class.name] = job
    end

    def resolve_dependencies
      return unless @dependencies

      @dependencies.each do |dependency|
        from = loaded_find_job(dependency[:from])
        to   = loaded_find_job(dependency[:to])
        to.incoming << dependency[:from]
        from.outgoing << dependency[:to]
      end
    end

    def loaded_find_job(id_or_name)
      @jobs_by_id[id_or_name] || @jobs_by_name[id_or_name]
    end

  end
end
