require 'spec_helper'
require 'pry'

describe "Workflows" do
  context "when all jobs finish successfuly" do
    it "marks workflow as completed" do
      flow = TestWorkflow.create(auto_destroy: false)
      flow.start!
      while ActiveJob::Base.queue_adapter.enqueued_jobs.any? do
        perform_one
      end

      flow = flow.reload
      expect(flow).to be_finished
      expect(flow).to_not be_failed
    end
  end

  context "when custom settings are provided" do
    context "auto destroy" do
      it "should runs the auto destroy when set for true" do
        flow = TestWorkflow.create(auto_destroy: true)
        flow.start!
        while ActiveJob::Base.queue_adapter.enqueued_jobs.any? do
          perform_one
        end

        database_flow = Kickstand::Workflow.where(id: flow.id).first
        expect(database_flow).to eq(nil)
      end
    end

    context "parallel execs" do
      it "should respect the number of parallel_executions provided" do
        flow = RepetitiveWorkflow.create(auto_destroy: false, parallel_execs: 1)
        flow.start!
        expect {
          perform_one #EnqueuerWorker
        }.to change{ActiveJob::Base.queue_adapter.enqueued_jobs.size}.from(1).to(2) # EnqueuerWorker and RepetitiveJob first
        expect(Kickstand::Worker).to have_jobs(flow.id, jobs_with_id(['RepetitiveJob']))

        ActiveJob::Base.queue_adapter.enqueued_jobs.clear
        another_flow = RepetitiveWorkflow.create(auto_destroy: false, parallel_execs: 2)
        another_flow.start!

        expect {
          perform_one # EnqueuerWorker
        }.to change{ActiveJob::Base.queue_adapter.enqueued_jobs.size}.from(1).to(3) # EnqueuerWorker, Prepare and RepetitiveJob first and second
      end
    end

    context "enqueuer_frequency" do
      it "should respect the enqueuer_frequency provided" do
        flow = TestWorkflow.create(auto_destroy: false, enqueuer_frequency: 10.seconds)
        flow.start!
        expect(Kickstand::EnqueuerWorker).to receive(:set).with(queue: "kickstand", wait: 10.seconds).and_call_original
        perform_one # EnqueuerWorker
      end

      it "should use the default when nothing is provided" do
        flow = TestWorkflow.create(auto_destroy: false)
        flow.start!
        expect(Kickstand::EnqueuerWorker).to receive(:set).with(queue: "kickstand", wait: 30.seconds).and_call_original
        perform_one # EnqueuerWorker
      end
    end
  end

  describe 'when one of the jobs fails initially' do
    class FailsThenSucceeds < Kickstand::Job
      def perform
        if FAIL_THEN_SUCCEED_SPY.foo == 'failure'
          raise NameError
        end
      end
    end

    class SecondChanceWorkflow < Kickstand::Workflow
      def configure
        prepare_job = run Prepare
        fails_then_suceeds_job = run FailsThenSucceeds, after: prepare_job
        run NormalizeJob, after: fails_then_suceeds_job
      end
    end

    context "background enqueuer" do
      it 'succeeds when the job retries' do
        FAIL_THEN_SUCCEED_SPY = double()
        allow(FAIL_THEN_SUCCEED_SPY).to receive(:foo).and_return('failure', 'success')

        flow = SecondChanceWorkflow.create(auto_destroy: false)
        flow.start!
        perform_one # EnqueuerWorker

        perform_one # Prepare
        job = job_by_name(flow, 'Prepare')
        expect(job).to be_finished
        perform_one # EnqueuerWorker

        perform_one # FailsThenSucceeds first attempt
        job = job_by_name(flow, 'FailsThenSucceeds')
        expect(job.failed?).to eq(true) # retrying

        # Retry the same job again, but this time succeeds
        perform_one # FailsThenSucceeds retry
        perform_one # EnqueuerWorker

        job = job_by_name(flow, 'NormalizeJob')
        perform_one
        perform_one # EnqueuerWorker
        perform_one

        flow = flow.reload
        expect(flow).to be_finished
        expect(flow).to_not be_failed
      end
    end
  end

  describe "runs the whole workflow in proper order" do
    context "background enqueuer" do
      it "runs the whole workflow in proper order" do
        flow = TestWorkflow.create
        flow.start!
        perform_one #EnqueuerWorker
        expect(Kickstand::Worker).to have_jobs(flow.id, jobs_with_id(['Prepare']))
        perform_one #PrepareJob

        expect(Kickstand::Worker).to have_no_jobs(flow.id, jobs_with_id(["FetchFirstJob", "FetchSecondJob"]))
        perform_one #EnqueuerWorker
        expect(Kickstand::Worker).to have_jobs(flow.id, jobs_with_id(["FetchFirstJob", "FetchSecondJob"]))

        perform_one #FetchFirstJob
        perform_one #FetchSecondJob

        expect(Kickstand::Worker).to have_no_jobs(flow.id, jobs_with_id(["PersistFirstJob"]))
        perform_one #EnqueuerWorker
        expect(Kickstand::Worker).to have_jobs(flow.id, jobs_with_id(["PersistFirstJob"]))

        perform_one #PersistFirstJob

        expect(Kickstand::Worker).to have_no_jobs(flow.id, jobs_with_id(["NormalizeJob"]))
        perform_one #EnqueuerWorker

        expect(Kickstand::Worker).to have_jobs(flow.id, jobs_with_id(["NormalizeJob"]))
        perform_one #NormalizeJob

        perform_one #EnqueuerWorker

        expect(ActiveJob::Base.queue_adapter.enqueued_jobs).to be_empty
      end
    end
  end

  context "payload" do
    class UpcaseJob < Kickstand::Job
      def perform
        data = { output: params[:input].upcase }
        output(data)
      end
    end

    it "passes payloads down the workflow by default" do
      class PrefixJob < Kickstand::Job
        def perform
          data = { output: params[:prefix].capitalize }
          output(data)
        end
      end

      class PrependJob < Kickstand::Job
        def perform
          prefix_job_payload = payloads.find { |j| j[:class] == 'PrefixJob'}
          prefix_job_string = prefix_job_payload[:output]["output"]
          upcase_job_payload = payloads.find { |j| j[:class] == 'UpcaseJob'}
          upcase_job_string = upcase_job_payload[:output]["output"]
          string = "#{prefix_job_string}: #{upcase_job_string}"
          output({output: string})
        end
      end

      class PayloadWorkflow < Kickstand::Workflow
        def configure
          upcase_job = run UpcaseJob, params: {input: "some text"}
          prefix_job = run PrefixJob, params: {prefix: "a prefix"}
          run PrependJob, after: [upcase_job, prefix_job]
        end
      end

      flow = PayloadWorkflow.create(auto_destroy: false)
      flow.start!

      perform_one #EnqueuerWorker
      perform_one #UpcaseJob
      perform_one #PrefixJob

      flow = flow.reload
      expect(flow.find_job("UpcaseJob").output_payload).to eq({"output" => "SOME TEXT"})
      expect(flow.find_job("PrefixJob").output_payload).to eq({"output" => "A prefix"})

      perform_one #EnqueuerWorker
      perform_one #PrependJob
      expect(flow.find_job("PrependJob").output_payload).to eq({"output" => "A prefix: SOME TEXT"})

      perform_one #EnqueuerWorker
      flow = flow.reload
      expect(flow).to be_finished
    end

    it "passes payloads from workflow that runs multiple same class jobs with nameized payloads" do
      flow = RepetitiveWorkflow.create(auto_destroy: false)
      flow.start!

      while ActiveJob::Base.queue_adapter.enqueued_jobs.any? do
        perform_one
      end

      flow = flow.reload

      first_job_output = { "output" => 'first' }
      first_job = flow.jobs.find { |job| job.output_payload == first_job_output }
      expect(first_job.output_payload).to eq(first_job_output)

      second_job_output = { "output" => 'second' }
      second_job = flow.jobs.find { |job| job.output_payload == second_job_output }
      expect(second_job.output_payload).to eq(second_job_output)

      third_job_output = { "output" => 'third' }
      third_job = flow.jobs.find { |job| job.output_payload == third_job_output }
      expect(third_job.output_payload).to eq(third_job_output)

      symmary_job_output = { "summary" => ['first', 'second', 'third'] }
      summary_job = flow.find_job("SummaryJob")
      expect(summary_job.output_payload).to eq(symmary_job_output)
    end
  end

  it "does not execute `configure` on each job for huge workflows" do
    INTERNAL_SPY = double('spy')
    INTERNAL_CONFIGURE_SPY = double('configure spy')
    expect(INTERNAL_SPY).to receive(:some_method).exactly(1005).times

    # One time when persisting, second time when reloading in the spec
    expect(INTERNAL_CONFIGURE_SPY).to receive(:some_method).exactly(1).times

    class SimpleJob < Kickstand::Job
      def perform
        INTERNAL_SPY.some_method
      end
    end

    class GiganticWorkflow < Kickstand::Workflow
      def configure
        INTERNAL_CONFIGURE_SPY.some_method

        dependents = []
        1000.times do
          dependents << run(SimpleJob)
        end

        5.times do
          run(SimpleJob, after: dependents)
        end
      end
    end

    flow = GiganticWorkflow.create(auto_destroy: false)
    flow.start!

    while ActiveJob::Base.queue_adapter.enqueued_jobs.any? do
      perform_one
    end

    flow = flow.reload
    expect(flow).to be_finished
    expect(flow).to_not be_failed
  end

  it 'executes job with multiple ancestors only once' do
    NO_DUPS_INTERNAL_SPY = double('spy')
    expect(NO_DUPS_INTERNAL_SPY).to receive(:some_method).exactly(1).times

    class FirstAncestor < Kickstand::Job
      def perform
      end
    end

    class SecondAncestor < Kickstand::Job
      def perform
      end
    end

    class FinalJob < Kickstand::Job
      def perform
        NO_DUPS_INTERNAL_SPY.some_method
      end
    end

    class NoDuplicatesWorkflow < Kickstand::Workflow
      def configure
        first_ancestor_job = run FirstAncestor
        second_ancestor_job = run SecondAncestor

        run FinalJob, after: [first_ancestor_job, second_ancestor_job]
      end
    end

    flow = NoDuplicatesWorkflow.create(auto_destroy: false)
    flow.start!

    6.times do
      perform_one
    end

    flow = flow.reload
    expect(flow).to be_finished
    expect(flow).to_not be_failed
  end
end
