require 'spec_helper'

describe Kickstand::Workflow do
  subject { TestWorkflow.create }

  describe "#initialize" do
    it "passes constructor arguments to the method" do
      klass = Class.new(Kickstand::Workflow) do
        def configure(*args)
          fetch_first_job = run FetchFirstJob
          run PersistFirstJob, after: fetch_first_job
        end
      end

      expect_any_instance_of(klass).to receive(:configure).with("arg1", "arg2")
      klass.create("arg1", "arg2")
    end

    it "correcly links tenant" do
      klass = Class.new(Kickstand::Workflow) do
        def configure(*args)
          link_tenant "organization", "my_org_id"
          fetch_first_job = run FetchFirstJob
          run PersistFirstJob, after: fetch_first_job
        end
      end
      klass.create("arg1", "arg2")
      workflow = Kickstand::Workflow.last
      expect(workflow.tenant).to eq("Organization: my_org_id")
    end

    it "passes options to the job" do
      klass = Class.new(Kickstand::Workflow) do
        def configure
          run FetchFirstJob,
            queue: :a,
            params: {a: "b"},
            include_payload: false,
            allowed_to_fail: true,
            retries: 4,
            timeout: 600
        end
      end
      klass.create
      job = Kickstand::Job.last
      expect(job.params).to eq(a: "b")
      expect(job.queue).to eq("a")
      expect(job.include_payload).to eq(false)
      expect(job.allowed_to_fail).to eq(true)
      expect(job.retries).to eq(4)
      expect(job.timeout).to eq(600)
    end

    it "applies default options" do
      klass = Class.new(Kickstand::Workflow) do
        def configure
          run FetchFirstJob, queue: :a
        end
      end
      klass.create
      job = Kickstand::Job.last
      expect(job.params).to eq({})
      expect(job.queue).to eq("a")
      expect(job.include_payload).to eq(true)
      expect(job.allowed_to_fail).to eq(false)
      expect(job.retries).to eq(1)
      expect(job.timeout).to eq(300)
    end
  end

  describe "create" do
    it "sets parallel_execs correctly when provided" do
      workflow = TestWorkflow.create(parallel_execs: 10)
      expect(workflow.parallel_execs).to eq(10) #assign
      database_workflow = Kickstand::Workflow.find(workflow.id)
      expect(database_workflow.parallel_execs).to eq(10) #persisted
    end

    it "sets parallel_execs as default when not provided" do
      workflow = TestWorkflow.create()
      expect(workflow.parallel_execs).to eq(500) #assign
      database_workflow = Kickstand::Workflow.find(workflow.id)
      expect(database_workflow.parallel_execs).to eq(500) #persisted
    end

    it "sets enqueuer_frequency correctly when provided" do
      workflow = TestWorkflow.create(enqueuer_frequency: 10.seconds)
      expect(workflow.enqueuer_frequency).to eq(10.seconds) #assign
      database_workflow = Kickstand::Workflow.find(workflow.id)
      expect(database_workflow.enqueuer_frequency).to eq(10.seconds) #persisted
    end

    it "sets enqueuer_frequency as default when not provided" do
      workflow = TestWorkflow.create()
      expect(workflow.enqueuer_frequency).to eq(30.seconds) #assign
      database_workflow = Kickstand::Workflow.find(workflow.id)
      expect(database_workflow.enqueuer_frequency).to eq(30.seconds) #persisted
    end

    it "sets timestamps in jobs correctly" do
      workflow = TestWorkflow.create()
      jobs = workflow.jobs.map(&:created_at).compact.length
      expect(jobs).to eq(workflow.jobs.length)
    end
  end

  describe "#delayed_created" do
    it "should allow the use of hash params when enqueuing a workflow asynchronously" do
      ParameterTestWorkflow.delayed_create({ something: 1 }, enqueuer_frequency: 55, parallel_execs: 10, auto_destroy: false)
      job = ActiveJob::Base.queue_adapter.enqueued_jobs.first # delayed_workflow_creator_worker
      perform_one
      expect(job["arguments"]).to match_array(["ParameterTestWorkflow", "{\"something\":1}", 55, 10, false, nil])

      perform_one # enqueuer
      job = ActiveJob::Base.queue_adapter.enqueued_jobs.first
      workflow_id = job["arguments"].first
      job_id = job["arguments"].last
      workflow = Kickstand::Workflow.find(workflow_id)
      job = workflow.find_job(job_id)
      expect(job.class).to be(Prepare)
      expect(Kickstand::Worker).to have_jobs(workflow.id, jobs_with_id(['Prepare']))
    end
  end

  describe "#mark_as_stopped" do
    it "marks workflow as stopped" do
      expect{ subject.mark_as_stopped }.to change{subject.stopped?}.from(false).to(true)
    end
  end

  describe "#mark_as_started" do
    it "removes stopped flag" do
      subject.stopped_at = DateTime.now
      expect{ subject.mark_as_started }.to change{subject.stopped?}.from(true).to(false)
    end
  end

  describe "#find_job" do
    it "finds job by its name" do
      workflow = TestWorkflow.create
      job = workflow.find_job("FetchFirstJob")
      expect(job).to be_instance_of(FetchFirstJob)
    end
  end

  describe "#run" do
    it "allows passing additional params to the job" do
      flow = RepetitiveWorkflow.create
      expect(flow.jobs.first.params).to eq ({ input: 'first' })
    end

    context "when graph is empty" do
      it "adds new job with the given class as a node" do
        flow = RepetitiveWorkflow.create
        expect(flow.jobs.first).to be_instance_of(RepetitiveJob)
      end
    end

    it "allows `after` to accept an array of jobs" do
      flow = TestWorkflow.create

      fetch_first_job = flow.find_job("FetchFirstJob")
      persist_first_job = flow.find_job("PersistFirstJob")

      expect(fetch_first_job.outgoing).to match_array([persist_first_job.id])
    end

    it "allows `before` to accept an array of jobs" do
      flow = TestWorkflow.create

      fetch_first_job = flow.find_job("FetchFirstJob")
      persist_first_job = flow.find_job("PersistFirstJob")

      expect(persist_first_job.incoming).to match_array([fetch_first_job.id])
    end

  end

  describe "#failed?" do
    context "when one of the jobs failed" do
      it "returns true" do
        subject.mark_as_failed
        expect(subject.failed?).to be_truthy
      end
    end

    context "when no jobs failed" do
      it "returns true" do
        expect(subject.failed?).to be_falsy
      end
    end
  end

  describe "#running?" do
    context "when no enqueued or running jobs" do
      it "returns false" do
        expect(subject.running?).to be_falsy
      end
    end

    context "when some jobs are running" do
      it "returns true" do
        subject.mark_as_started
        expect(subject.running?).to be_truthy
      end
    end
  end

  describe "#finished?" do
    it "returns true if finished_at is filled" do
      subject.finished_at = Time.now
      expect(subject.finished?).to be_truthy
    end
  end
end
