$LOAD_PATH.unshift File.expand_path("../../lib", __FILE__)
require 'pry-byebug'

# Models to test against in spec
#MODELS = File.join(File.dirname(__FILE__), "models")

require "pry"
require "database_cleaner-mongoid"
require "timecop"
require "kickstand"

Mongoid.configure do |config|
  config.connect_to "kickstand"
  config.logger = nil
  config.clients.default = {
    hosts: ['127.0.0.1:27017'],
    database: 'my_db',
  }
end

KICKSTANDFILE = Pathname.new(__FILE__).parent.join("Kickstandfile")
REDIS_URL = "redis://localhost:6379/12"

Mongo::Logger.level = ::Logger::FATAL
ActiveJob::Base.queue_adapter = :test
ActiveJob::Base.logger = nil

module KickstandHelpers
  def redis
    @redis ||= Redis.new(url: REDIS_URL)
  end

  def perform_one
    job = ActiveJob::Base.queue_adapter.enqueued_jobs.first
    if job
      job[:job].new.perform(*job[:args])
      ActiveJob::Base.queue_adapter.performed_jobs << job
      ActiveJob::Base.queue_adapter.enqueued_jobs.shift
    end
  end

  def perform_enqueuer
    job = ActiveJob::Base.queue_adapter.enqueued_jobs.find { |job| job[:job] == Kickstand::EnqueuerWorker }
    if job
      job[:job].new.perform(*job[:args])
      ActiveJob::Base.queue_adapter.performed_jobs << job
      ActiveJob::Base.queue_adapter.enqueued_jobs.shift
    end
  end

  def find_job_by_name(workflow_id, job_name)
    ActiveJob::Base.queue_adapter.enqueued_jobs.select { |job| job[:job] == actual }
  end

  def jobs_with_id(jobs_array)
    jobs_array.map {|job_name| job_with_id(job_name) }
  end

  def job_with_id(job_name)
    /#{job_name}|(?<identifier>.*)/
  end

  def job_by_name(workflow, job_name)
    reloaded = workflow.reload
    reloaded.jobs.find { |job| job.class.name.to_s == job_name }
  end
end

RSpec.configure do |config|
  config.include ActiveJob::TestHelper
  config.include KickstandHelpers

  config.mock_with :rspec do |mocks|
    mocks.verify_partial_doubles = true
  end

  config.before(:each) do
    Kickstand.configure do |config|
      config.redis_opts = { url: REDIS_URL, :connect_timeout => 1, :read_timeout => 10, :write_timeout => 60 }
      config.kickstandfile   = KICKSTANDFILE
      config.on_job_failure = ->(job, error) { puts "Job #{ job.name } failed, due to #{ error.to_s }" }
      config.on_workflow_failure = ->(workflow) { puts("Workflow #{ workflow.class } failed") }
      config.on_job_retry = ->(workflow, job) { puts("#{ workflow.class } - #{ workflow.id } - #{ job.name } retrying") }
      config.on_job_timeout = ->(workflow, job) { puts("#{ workflow.class } - #{ workflow.id } - #{ job.name } - timeout") }
      config.on_job_lost = ->(workflow, job) { puts("#{ workflow.class } - #{ workflow.id } - #{ job.name } lost") }
      config.on_job_enqueue = ->(workflow, job) { puts("#{ workflow.class } - #{ workflow.id } - #{ job.name } enqueued") }
      config.on_workflow_persist = ->(workflow) { puts("#{ workflow.class } - #{ workflow.id } persisted") }
      config.on_workflow_start = ->(workflow) { puts("#{ workflow.class } - #{ workflow.id } started") }
      config.on_workflow_finish = ->(workflow) { puts("#{ workflow.class } - #{ workflow.id } finished") }
      config.on_workflow_auto_destroy = ->(workflow) { puts("#{ workflow.class } - #{ workflow.id } auto destroyed") }
    end
  end

  config.before(:all) do
    Kickstand::Job.all.delete_all
    Kickstand::Workflow.all.delete_all
  end

  config.before(:each) do
    DatabaseCleaner.start
    clear_enqueued_jobs
    clear_performed_jobs
  end

  config.after(:each) do
    DatabaseCleaner.clean
    clear_enqueued_jobs
    clear_performed_jobs
    redis.flushdb
  end

  config.after(:all) do
    clear_enqueued_jobs
    clear_performed_jobs
    redis.flushdb
    DatabaseCleaner.clean
  end
end

class Prepare < Kickstand::Job; end
class FetchFirstJob < Kickstand::Job; end
class FetchSecondJob < Kickstand::Job; end
class PersistFirstJob < Kickstand::Job; end
class PersistSecondJob < Kickstand::Job; end
class NormalizeJob < Kickstand::Job; end
class BobJob < Kickstand::Job; end

class FailJob < Kickstand::Job
  def perform
    raise StandardError, "failed"
  end
end

class RepetitiveJob < Kickstand::Job
  def perform
    output({ output: params[:input] })
  end
end

class SummaryJob < Kickstand::Job
  def perform
    outputs = payloads.map { |payload| payload[:output]["output"] }
    data = { summary: outputs }
    output data
  end
end

class TestWorkflow < Kickstand::Workflow
  def configure
    prepare_job = run Prepare

    normalize_job = run NormalizeJob

    fetch_first_job = run FetchFirstJob,   after: prepare_job
    run FetchSecondJob,  after: prepare_job, before: normalize_job

    run PersistFirstJob, after: fetch_first_job, before: normalize_job
  end
end

class FailRetryWorkflow < Kickstand::Workflow
  def configure
    run FailJob, allowed_to_fail: true, retries: 2
  end
end

class ParameterTestWorkflow < Kickstand::Workflow
  def configure(param)
    run Prepare if param
  end
end

class RepetitiveWorkflow < Kickstand::Workflow
  def configure
    jobs = []
    jobs << run(RepetitiveJob, params: {input: 'first'})
    jobs << run(RepetitiveJob, params: {input: 'second'})
    jobs << run(RepetitiveJob, params: {input: 'third'})
    run SummaryJob, after: jobs
  end
end

RSpec::Matchers.define :have_jobs do |flow, jobs|
  match do |actual|
    expected = jobs.map do |job|
      hash_including(args: include(flow, job))
    end
    filtered_jobs = ActiveJob::Base.queue_adapter.enqueued_jobs.select { |job| job[:job] == actual }
    expect(filtered_jobs).to match_array(expected)
  end

  failure_message do |actual|
    "expected queue to have #{jobs}, but instead has: #{ActiveJob::Base.queue_adapter.enqueued_jobs.map{ |j| j[:args][1]}}"
  end
end

RSpec::Matchers.define :have_no_jobs do |flow, jobs|
  match do |actual|
    expected = jobs.map do |job|
      hash_including(args: include(flow, job))
    end
    filtered_jobs = ActiveJob::Base.queue_adapter.enqueued_jobs.select { |job| job[:job] == actual }
    expect(filtered_jobs).not_to match_array(expected)
  end

  failure_message do |actual|
    "expected queue to have no #{jobs}, but instead has: #{ActiveJob::Base.queue_adapter.enqueued_jobs.map{ |j| j[:args][1]}}"
  end
end
